#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image


def scale_image(
    scale: tuple,
    image: Optional[np.ndarray] = None,
    mask: Optional[np.ndarray] = None,
    nn=False,
    device="cpu",
) -> Union[
    Tuple[np.ndarray, np.ndarray, int], Tuple[np.ndarray, int], Tuple[None, int]
]:
    """
    Scale image module using torch GPU interpolation

    Expect 2D (XY/YX), 3D (ZYX)

    Args:
        image (np.ndarray, Optional): image data
        mask (np.ndarray, Optional): Optional binary mask image data
        scale (tuple): scale value for image
        nn (bool):
        device (str):
    """
    dim = 1
    type_i = image.dtype
    type_m = None
    if mask is not None:
        type_m = mask.dtype

    if len(scale) == 3:
        scale = tuple([scale[0], scale[1], scale[2]])
    else:
        scale = tuple([scale[0], scale[1]])

    if nn:
        image = nn_scaling(img=image, scale=scale, dtype=type_i, device=device)
        return image, dim

    if image is not None:
        if not np.all(scale == image.shape):
            if scale[0] > image.shape[0]:
                if image.ndim == 3 and image.shape[2] != 3:  # 3D with Gray
                    image = area_scaling(
                        img=image, scale=scale, dtype=type_i, device=device
                    )
                else:
                    image = linear_scaling(
                        img=image, scale=scale, dtype=type_i, device=device
                    )
            else:
                image = pil_LANCZOS(
                    img=image.astype(np.float32), scale=scale, dtype=type_i
                )

    if mask is not None:
        if not np.all(scale == mask.shape):
            if scale[0] > mask.shape[0]:
                mask = linear_scaling(img=mask, scale=scale, dtype=type_m)
            else:
                mask = pil_LANCZOS(img=mask.astype(np.int16), scale=scale, dtype=type_m)

    if image is not None and mask is not None:
        return image, mask, dim
    elif image is None and mask is not None:
        return mask, dim
    else:
        return image, dim


def nn_scaling(
    img: np.ndarray, scale: tuple, dtype: np.dtype, device="cpu"
) -> np.ndarray:
    img = torch.from_numpy(img)[None, None, ...].to(device)
    img = F.interpolate(img, size=scale, mode="nearest")[0, 0, ...]

    return img.cpu().detach().numpy()


def pil_LANCZOS(img: np.ndarray, scale: tuple, dtype: np.dtype) -> np.ndarray:
    """
    Scale a 3D image by first down-sampling in the XY direction and then in the Z direction.

    Args:
        img: image array.
        scale: Scale array size.
        dtype: Output dtype for scale array.

    Returns:
        no.ndarray: Up or Down scale 3D array.
    """
    if len(scale) == 3:
        new_depth, new_height, new_width = scale
        scaled = []

        # Down-sample each slice in the XY direction
        for i in range(img.shape[0]):
            image = Image.fromarray(img[i, ...])
            scaled_image = image.resize((new_width, new_height), Image.LANCZOS)
            scaled.append(np.asarray(scaled_image, dtype=dtype))

        # Convert list to 3D array
        img = np.array(scaled, dtype=dtype)

        # Down-sample each slice in the Z direction if 3D
        if len(scale) == 3:
            scaled = np.zeros(scale, dtype=dtype)

            for i in range(img.shape[2]):
                image = Image.fromarray(img[..., i])
                scaled[:, :, i] = np.asarray(
                    image.resize((new_height, new_depth), Image.LANCZOS), dtype=dtype
                )
            img = scaled.astype(dtype)
    else:
        new_height, new_width = scale
        img = Image.fromarray(img)
        img = np.asarray(
            img.resize((new_width, new_height), Image.LANCZOS), dtype=dtype
        )

    return img.astype(dtype)


def linear_scaling(
    img: np.ndarray, scale: tuple, dtype: np.dtype, device="cpu"
) -> np.ndarray:
    """
    Scaling of 2D/3D array using trilinear method from pytorch

    Args:
        img: image array.
        scale: Scale array size.
        dtype: Output dtype for scale array.

    Returns:
        no.ndarray: Up or Down scale 3D array.
    """
    if img.ndim == 3:
        current_depth, current_height, current_width = (
            img.shape[0],
            img.shape[1],
            img.shape[2],
        )
        final_depth, final_height, final_width = scale

        img = torch.from_numpy(img[None, None, :]).to(device).type(torch.float)

        # Calculate the number of steps required to reach the final scale by halving/doubling
        # Use logarithm base 2 since scaling is done by 2x at each step
        height_steps = int(abs(torch.log2(torch.tensor(final_height / current_height))))
        width_steps = int(abs(torch.log2(torch.tensor(final_width / current_width))))
        depth_steps = int(abs(torch.log2(torch.tensor(final_depth / current_depth))))

        # Perform scaling in steps
        for _ in range(max(height_steps, width_steps, depth_steps)):
            # Calculate intermediate scale
            new_height = int(
                current_height * (2 if current_height < final_height else 0.5)
            )
            new_width = int(current_width * (2 if current_width < final_width else 0.5))
            new_depth = int(current_depth * (2 if current_depth < final_depth else 0.5))

            # Stop if the desired scale is reached or exceeded
            if (
                new_height >= final_height
                and new_width >= final_width
                and new_depth >= final_depth
            ):
                break

            # Resize image
            img = F.interpolate(
                img,
                size=(new_height, new_width, new_depth),
                mode="trilinear",
                align_corners=False,
            )

            # Update current dimensions
            current_depth, current_height, current_width = (
                new_depth,
                new_height,
                new_width,
            )

            # Stop if the desired scale is reached or exceeded
            if (
                current_height >= final_height
                and current_width >= final_width
                and current_depth >= final_depth
            ):
                break

        img = F.interpolate(img, size=scale, mode="trilinear", align_corners=False)
    else:
        current_height, current_width = img.shape[0], img.shape[1]
        final_height, final_width = scale

        img = torch.from_numpy(img[None, None, :]).to(device).type(torch.float)

        # Calculate the number of steps required to reach the final scale by halving/doubling
        # Use logarithm base 2 since scaling is done by 2x at each step
        height_steps = int(abs(torch.log2(torch.tensor(final_height / current_height))))
        width_steps = int(abs(torch.log2(torch.tensor(final_width / current_width))))

        # Perform scaling in steps
        if max(height_steps, width_steps) > 0:
            for _ in range(max(height_steps, width_steps)):
                # Calculate intermediate scale
                new_height = int(
                    current_height * (2 if current_height < final_height else 0.5)
                )
                new_width = int(
                    current_width * (2 if current_width < final_width else 0.5)
                )

                # Stop if the desired scale is reached or exceeded
                if new_height >= final_height and new_width >= final_width:
                    break

                # Resize image
                img = F.interpolate(
                    img,
                    size=(new_height, new_width),
                    mode="bilinear",
                    align_corners=False,
                )
                current_height, current_width = new_height, new_width

        img = F.interpolate(img, size=scale, mode="bilinear", align_corners=False)
    return img.detach().numpy()[0, 0, :].astype(dtype)


def area_scaling(
    img: np.ndarray, scale: tuple, dtype: np.dtype, device="cpu"
) -> np.ndarray:
    """
    Scaling of 3D array using area method from pytorch

    Args:
        img: 3D array.
        scale: Scale array size.
        dtype: Output dtype for scale array.

    Returns:
        no.ndarray: Up or Down scale 3D array.
    """

    size_Z = [scale[0], img.shape[1], img.shape[2]]
    image_scale_Z = np.zeros(size_Z, dtype=dtype)

    # Scale Z axis
    for i in range(img.shape[2]):
        df_img = torch.from_numpy(img[:, :, i]).to(device).type(torch.float)

        image_scale_Z[:, :, i] = (
            F.interpolate(
                df_img[None, None, :], size=[int(s) for s in size_Z[:2]], mode="area"
            )
            .cpu()
            .detach()
            .numpy()[0, 0, :]
            .astype(dtype)
        )

    # Scale XY axis
    img = np.zeros(scale, dtype=dtype)
    for i in range(scale[0]):
        df_img = torch.from_numpy(image_scale_Z[i, :]).to(device).type(torch.float)
        img[i, :] = (
            F.interpolate(
                df_img[None, None, :], size=[int(s) for s in scale[1:]], mode="area"
            )
            .cpu()
            .detach()
            .numpy()[0, 0, :]
            .astype(dtype)
        )

    return img


def number_of_features_per_level(channel_scaler: int, num_levels: int) -> list:
    """
    Compute list of output channels for CNN.

    Features = channel_scaler * 2^k
        where:
        - k is layer number

    Args:
        channel_scaler (int): Number of initial input channels for CNN.
        num_levels (int): Number of channels from max_number_of_conv_layer().

    Returns:
        list: List of output channels.
    """
    return [channel_scaler * 2**k for k in range(num_levels)]


def max_number_of_conv_layer(
    img=None,
    input_volume=64,
    max_out=8,
    kernel_size=3,
    padding=1,
    stride=1,
    pool_size=2,
    pool_stride=2,
    first_max_pool=False,
) -> int:
    """
    Calculate maximum possible number of layers given image size.

    Based on the torch input automatically select number of convolution blocks,
    based on a standard settings.

        I = [(W - K + 2*P) / S] + 1
        Out_size = [(I - F) / S] + 1
        - W is the input volume
        - K is the Kernel size
        - P is the padding
        - S is the stride
        - F is the max pooling kernel size

    Args:
        img (np.ndarray, Optional): Tensor data from which size of is calculated.
        input_volume (int): Size of the multiplayer for the convolution.
        max_out (int): Maximal output dimension after max_pooling.
        kernel_size (int): Kernel size for the convolution.
        padding (int): Padding size for the convolution blocks.
        stride: (int) Stride for the convolution blocks.
        pool_size (int) Max Pooling kernel size.
        pool_stride (int): Max Pooling stride size.
        first_max_pool (bool): If first CNN block has max pooling.

    Returns:
        int: Maximum number of CNN layers.
    """
    if img is not None:
        # In case of anisotropic image select smallest size
        # to calculate max conv layers
        if len(img.shape) == 5:
            size = np.array(list(img.shape[2:]))
        elif len(img.shape) == 4:
            size = np.array(list(img.shape[1:]))
        else:
            size = np.array(list(img.shape))

        size = size[size.argmin()]
        input_volume = size

    max_layers = 0

    while input_volume >= max_out:
        img_window = ((input_volume - kernel_size + 2 * padding) / stride) + 1
        input_volume = ((img_window - pool_size) / pool_stride) + 1
        max_layers += 1

    if first_max_pool:
        max_layers -= 1

    return max_layers


def normalize_image(image: np.ndarray) -> np.ndarray:
    """
    Simple image data normalizer between 0,1.

    Args:
        image: Image data set.

    Returns:
        np.ndarray: Normalized image between 0 and 1 values.
    """
    image_min = np.min(image)
    image_max = np.max(image)

    if image_min == 0 and image_max == 1:
        return image

    if image_min == 0:
        image = np.where(image > image_min, 1, 0)
    elif image_max == 0:
        image = np.where(image < image_max, 1, 0)

    return image


def check_model_dict(model_dict: dict) -> dict:
    """
    Check and rebuild model structure dictionary to ensure back-compatibility.

    Args:
        model_dict (dict): Model structure dictionary.

    Returns:
        dict: Standardize model structure dictionary.
    """
    new_dict = {}

    for key, value in model_dict.items():
        if key.endswith("type"):
            new_dict["cnn_type"] = value
        if key.endswith("cation"):
            new_dict["classification"] = value
        if key.endswith("_in") or key.startswith("in_"):
            new_dict["in_channel"] = value
        if key.endswith("_out") or key.startswith("out_"):
            new_dict["out_channel"] = value
        if key.endswith("size"):
            new_dict["img_size"] = value
        if key.endswith("dropout"):
            new_dict["dropout"] = value
        if key.endswith("layers"):
            new_dict["num_conv_layers"] = value
        if key.endswith("scaler") or key.endswith("multiplayer"):
            new_dict["conv_scaler"] = value
        if key.endswith("v_kernel"):
            new_dict["conv_kernel"] = value
        if key.endswith("padding"):
            new_dict["conv_padding"] = value
        if key.endswith("l_kernel"):
            new_dict["maxpool_kernel"] = value
        if key.endswith("components"):
            new_dict["layer_components"] = value
        if key.endswith("features"):
            new_dict["attn_features"] = value
        if key.endswith("group"):
            new_dict["num_group"] = value

    return new_dict



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import numpy as np
import torch

from tardis_em.utils.metrics import calculate_f1
from tardis_em.utils.trainer import BasicTrainer


class CNNTrainer(BasicTrainer):
    """
    GENERAL CNN TRAINER
    """

    def __init__(self, **kwargs):
        super(CNNTrainer, self).__init__(**kwargs)

    def _train(self):
        """
        Run model training.
        """
        # Update progress bar
        self._update_progress_bar(loss_desc="Training: (loss 1.000)", idx=0)

        # Run training for CNN model
        for idx, (i, m) in enumerate(self.training_DataLoader):
            """Mid-training eval"""
            self._mid_training_eval(idx=idx)

            """Training"""
            i, m = i.to(self.device), m.to(self.device)
            self.optimizer.zero_grad()

            if self.classification:
                i, _ = self.model(i)  # one forward pass
            else:
                i = self.model(i)  # one forward pass

            # Back-propagate
            loss = self.criterion(i, m)

            loss.backward()  # one backward pass
            self.optimizer.step()  # update the parameters

            # Store training loss metric
            loss_value = loss.item()
            self.training_loss.append(loss_value)

            # Store and update learning rate
            if self.lr_scheduler:
                self.lr = self.optimizer.get_lr_scale()
            self.learning_rate.append(self.lr)

            # Update progress bar
            self._update_progress_bar(
                loss_desc=f"Training: (loss {loss_value:.4f};" f" LR: {self.lr:.5f})",
                idx=idx,
            )

    def _validate(self):
        """
        Test model against validation dataset.
        """
        valid_losses = []
        accuracy_mean = []
        precision_mean = []
        recall_mean = []
        F1_mean = []
        threshold_mean = []

        SM = torch.nn.Softmax(1)

        for idx, (img, mask) in enumerate(self.validation_DataLoader):
            img, mask = img.to(self.device), mask.to(self.device)

            with torch.no_grad():
                img = self.model(img)
                loss = self.criterion(img, mask)

                if img.shape[1] != 1:
                    img = SM(img)[0, :, :].flatten()
                    mask = mask[0, :, :].flatten()
                else:
                    img = torch.sigmoid(img)[0, 0, :]
                    mask = mask[0, 0, :]

                img = np.where(img.cpu().detach().numpy() >= 0.5, 1, 0)
                mask = mask.cpu().detach().numpy()
                acc, prec, recall, f1 = calculate_f1(
                    logits=img, targets=mask, best_f1=False
                )

                # Avg. precision score
                valid_losses.append(loss.item())
                accuracy_mean.append(acc)
                precision_mean.append(prec)
                recall_mean.append(recall)
                F1_mean.append(f1)
                threshold_mean.append(0.5)
                valid = (
                    f"Validation: (loss {loss.item():.4f} "
                    f"Prec: {prec:.2f} Rec: {recall:.2f} F1: {f1:.2f})"
                )

                # Update progress bar
                self._update_progress_bar(loss_desc=valid, idx=idx, train=False)

        # Reduce eval. metric with mean
        self.validation_loss.append(np.mean(valid_losses))
        self.accuracy.append(np.mean(accuracy_mean))
        self.precision.append(np.mean(precision_mean))
        self.recall.append(np.mean(recall_mean))
        self.threshold.append(np.mean(threshold_mean))
        self.f1.append(np.mean(F1_mean))

        # Check if average evaluation loss dropped
        self.early_stopping(f1_score=np.mean(F1_mean))



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import numpy as np
import os
import click


def tardis_helper(func: str, dir_=None):
    from tardis_em.utils.export_data import to_am

    if func == "csv_am":
        px = click.prompt(
            "Detected .tif files, please provide pixel size "
            "(It will be used for all .tif images):",
            type=float,
        )

        csv_files = [f for f in os.listdir(dir_) if f.endswith(".csv")]
        for i in csv_files:
            try:
                coord = np.genfromtxt(os.path.join(dir_, i), delimiter=",")
                if isinstance((coord[0, 0]), int) or isinstance((coord[0, 0]), float):
                    if coord.shape[1] == 3:
                        coord = np.vstack(
                            (coord, np.repeat(0, len(coord)).reshape(-1, 1))
                        )
                    to_am(coord, px, os.path.join(dir_, i[:-4] + ".csv"))
            except:
                pass
    elif func == "am_csv":
        from tardis_em.utils.load_data import ImportDataFromAmira

        # find a all .am files
        am_files = [f for f in os.listdir(dir_) if f.endswith(".am")]

        for i in am_files:
            try:
                am_import = ImportDataFromAmira(os.path.join(dir_, i))
                if am_import is not None:
                    am_import = am_import.get_segmented_points()
                    np.savetxt(
                        os.path.join(dir_, i[:-3] + ".csv"),
                        am_import,
                        delimiter=",",
                    )
            except:
                pass



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import asyncio
import boto3
import pandas as pd
import torch
import tifffile.tifffile as tif
import numpy as np
import json
import time

from os import mkdir, remove
from os.path import join, isdir, isfile
from botocore import UNSIGNED
from botocore.config import Config
from cryoet_data_portal import Client, Tomogram
from s3transfer import S3Transfer
from shutil import rmtree, move

from tardis_em.cnn.data_processing.stitch import StitchImages
from tardis_em.cnn.data_processing.trim import trim_with_stride
from tardis_em.cnn.datasets.dataloader import PredictionDataset
from tardis_em.cnn.utils.utils import scale_image
from tardis_em.dist_pytorch.datasets.patches import PatchDataSet
from tardis_em.dist_pytorch.utils.build_point_cloud import BuildPointCloud
from tardis_em.dist_pytorch.utils.segment_point_cloud import PropGreedyGraphCut
from tardis_em.utils.load_data import load_image, mrc_read_header
from tardis_em.utils.logo import TardisLogo, print_progress_bar
from tardis_em.utils.device import get_device
from tardis_em.utils.predictor import Predictor
from tardis_em.utils.normalization import RescaleNormalize, MeanStdNormalize
from tardis_em.utils.export_data import to_mrc
from tardis_em.utils.spline_metric import sort_by_length


async def get_from_aws(url):
    def download_file():
        s3 = boto3.client("s3", config=Config(signature_version=UNSIGNED))
        s3.download_file("cryoet-data-portal-public", url, url.split("/")[-1])

        print(f"Downloaded: {url.split('/')[-1]}")

    loop = asyncio.get_running_loop()
    # Run the download_file function in a separate thread
    await loop.run_in_executor(None, download_file)


async def upload_to_aws(
    aws_key, aws_secret, bucket, id_, data_name, file_name, remove_=True
):
    def upload_file():
        if data_name is None:
            client = boto3.client(
                "s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret
            )
            transfer = S3Transfer(client)
            transfer.upload_file(
                file_name,
                bucket,
                join("robert_kiewisz_tardis_01_2024", id_, file_name),
            )
        try:
            client = boto3.client(
                "s3", aws_access_key_id=aws_key, aws_secret_access_key=aws_secret
            )
            transfer = S3Transfer(client)
            transfer.upload_file(
                file_name,
                bucket,
                join("robert_kiewisz_tardis_01_2024", id_, data_name, file_name),
            )
            if remove_:
                remove(file_name)
            else:
                move(file_name, join("TARDIS", file_name))

            print(f"Uploaded: {file_name}")
        except Exception as e:
            print(f"Upload fail: {file_name}, Error: {str(e)}")

    loop = asyncio.get_running_loop()
    try:
        # Run the upload_file function in a separate thread
        await loop.run_in_executor(None, upload_file)
    except Exception as e:
        print(f"Upload fail: {file_name}, Error: {str(e)}")


class ProcessTardisForCZI:
    def __init__(self, aws: list, allocate_gpu: list, predict="Membrane"):
        self.upload_tasks = []

        self.aws_key, self.aws_secret, self.bucket = aws

        self.device_cnn = get_device(allocate_gpu[0])
        self.device_dist = get_device(allocate_gpu[1])

        self.predict = predict
        self.normalize_px = 15 if self.predict == "Membrane" else 25
        self.image_stitcher = StitchImages()
        self.post_processes = BuildPointCloud()

        self.output = join("temp", "Predictions")
        self.patches = join("temp", "Patches")
        self.results = "TARDIS"

        # Normalize histogram
        self.normalize = RescaleNormalize(clip_range=(1, 99))
        self.mean_std = MeanStdNormalize()

        self.terminal = TardisLogo()
        self.terminal(title="Predict CIZ-Cryo-EM data-port datasets")

        self.last_dist_task = None

        # Load CNN model
        self.model_cnn = Predictor(
            checkpoint=None,
            network="unet",
            subtype="32",
            model_type=(
                "membrane_3d" if self.predict == "Membrane" else "microtubules_3d"
            ),
            img_size=128,
            sigmoid=False,
            device=self.device_cnn,
        )

        self.model_dist = Predictor(
            checkpoint=None,
            network="dist",
            subtype="triang",
            model_type="3d" if self.predict == "Membrane" else "2d",
            device=self.device_dist,
        )

        self.patch_pc = PatchDataSet(max_number_of_points=1000, graph=False)
        self.segment_pc = PropGreedyGraphCut(
            threshold=0.5, connection=4 if self.predict == "Membrane" else 2
        )

    @staticmethod
    def clean_temp():
        if isdir("temp"):
            rmtree("temp")

    def build_temp(self):
        """
        Standard set-up for creating new temp dir for cnn prediction.

        Args:
            dir_ (str): Directory where folder will be build.
        """
        if not isdir(self.results):
            mkdir(self.results)

        self.clean_temp()

        mkdir("temp")
        mkdir(self.patches)
        mkdir(self.output)

    def pre_process(self, data, px, name):
        data = self.normalize(self.mean_std(data)).astype(np.float32)
        tif.imwrite(join(self.results, name + ".tif"), data.sum(0).astype(np.float32))

        if not data.min() >= -1 or not data.max() <= 1:  # Image not between in -1 and 1
            if data.min() >= 0 and data.max() <= 1:
                data = (data - 0.5) * 2  # shift to -1 - 1
            elif data.min() >= 0 and data.max() <= 255:
                data = data / 255  # move to 0 - 1
                data = (data - 0.5) * 2  # shift to -1 - 1

        scale_factor = px / self.normalize_px
        org_shape = data.shape
        scale_shape = np.multiply(org_shape, scale_factor).astype(np.int16)
        scale_shape = [int(i) for i in scale_shape]

        return data, scale_factor, org_shape, scale_shape

    def predict_cnn(self, dataloader):
        for j in range(len(dataloader)):
            input_, name = dataloader.__getitem__(j)
            input_ = self.model_cnn.predict(input_[None, :], rotate=False)
            tif.imwrite(join(self.output, f"{name}.tif"), input_)

    def predict_dist(self, coords_df):
        x = []
        with torch.no_grad():
            for i in coords_df:
                x.append(self.model_dist.predict(x=i[None, :], y=None))

        return x

    @staticmethod
    def prediction_stat(name: list, pc: np.ndarray):
        if isfile("stat.json"):
            with open("stat.json", "r") as file:
                data = json.load(file)
        else:
            data = {}

        try:
            name_ = name[0]
            data[name_] = {}
            data[name_]["Pixel_Size"] = name[1]
            data[name_]["Segmentation_time"] = name[2]
            data[name_]["Number_of_Points"] = len(pc)
            data[name_]["Number_of_Instances"] = len(np.unique(pc[:, 0]))
            data[name_]["Size_of_Each_Instance"] = [
                len(pc[pc[:, 0] == x, :]) for x in np.unique(pc[:, 0])
            ]

            size_ = np.array(data[name_]["Size_of_Each_Instance"])
            data[name_]["Contain_Membrane"] = 1 if np.any(size_ > 1000) else 0
            if data[name_]["Contain_Membrane"] == 0:
                data[name_]["Maybe_Contain_Membrane"] = np.any(size_ > 250)
            else:
                data[name_]["Maybe_Contain_Membrane"] = 1

            with open("stat.json", "w") as file:
                json.dump(data, file, indent=4)
        except TypeError:
            name_ = name[0]
            data[name_] = {}
            data[name_]["Pixel_Size"] = name[1]
            data[name_]["Segmentation_time"] = name[2]
            data[name_]["Number_of_Points"] = "NaN"
            data[name_]["Number_of_Instances"] = "NaN"
            data[name_]["Size_of_Each_Instance"] = "NaN"

            data[name_]["Contain_Membrane"] = "NaN"
            data[name_]["Maybe_Contain_Membrane"] = "NaN"

            with open("stat.json", "w") as file:
                json.dump(data, file, indent=4)

    async def cnn_prediction(
        self, data: list, px: float, header, name: str, progress: tuple
    ):
        """Build temp dir"""
        self.build_temp()

        """Load and prepare dataset for prediction"""
        self.terminal(
            text_1="TARDIS segmentation:",
            text_2="Preprocess data...",
            text_3=f"Dataset: [{name}]",
            text_4=f"Pixel_size: {px}",
            text_6=print_progress_bar(progress[0], progress[1]),
        )

        data, scale_factor, org_shape, scale_shape = data
        trim_with_stride(
            image=data,
            scale=scale_shape,
            trim_size_xy=128,
            trim_size_z=128,
            output=self.patches,
            image_counter=0,
            clean_empty=False,
            stride=round(128 * 0.125),
            device=self.device_cnn,
        )

        """TARDIS - Semantic Segmentation"""
        self.terminal(
            text_1="TARDIS segmentation:",
            text_2="Semantic Segmentation...",
            text_3=f"Dataset: [{name}]",
            text_4=f"Pixel_size: {px}",
            text_6=print_progress_bar(progress[0], progress[1]),
        )

        # CNN predict
        self.name = name
        self.px = px
        self.progress = (progress[0], progress[1])
        self.predict_cnn(
            dataloader=PredictionDataset(join(self.patches, "imgs")),
        )

        # Post process
        data = self.image_stitcher(image_dir=self.output, mask=False, dtype=np.float32)[
            : scale_shape[0], : scale_shape[1], : scale_shape[2]
        ]
        data, _ = scale_image(
            image=data, scale=org_shape, nn=False, device=self.device_cnn
        )
        data = torch.sigmoid(torch.from_numpy(data)).cpu().detach().numpy()

        data = np.where(data > 0.25, 1, 0).astype(np.uint8)
        to_mrc(data, px, name + "_semantic.mrc", header)  # To upload back to AWS
        tif.imwrite(
            join(self.results, name + "_semantic.tif"), data.sum(0).astype(np.uint8)
        )  # Store for NYSBC

        self.clean_temp()

        return data

    async def dist_prediction(
        self,
        data,
        px: float,
        start_time: float,
        name: str,
        folder_name: str,
        progress: tuple,
    ):
        data = await data

        self.upload_tasks += [
            asyncio.create_task(
                upload_to_aws(
                    self.aws_key,
                    self.aws_secret,
                    self.bucket,
                    folder_name,
                    name,
                    name + "_semantic.mrc",
                )
            )
        ]

        self.terminal(
            text_1="TARDIS segmentation:",
            text_2="Instance Segmentation...",
            text_3=f"Dataset: [{name}]",
            text_4=f"Pixel_size: {px}",
            text_6=print_progress_bar(progress[0], progress[1]),
        )

        _, pc_ld = BuildPointCloud().build_point_cloud(
            image=data,
            down_sampling=5,
            as_2d=True if self.predict == "Membrane" else False,
        )
        if len(pc_ld) > 100:
            coords_df, _, output_idx, _ = self.patch_pc.patched_dataset(coord=pc_ld)
            x = self.predict_dist(coords_df)

            try:
                pc = self.segment_pc.patch_to_segment(
                    graph=x,
                    coord=pc_ld,
                    idx=output_idx,
                    sort=False if self.predict == "Membrane" else True,
                    prune=25 if self.predict == "Membrane" else 10,
                )
            except ValueError:
                pc = np.zeros((1, 4))  # Empty Instance, no instance found
        else:
            pc = np.zeros((0, 4))

        end_time = np.round(time.time() - start_time, 0) / 60  # Minutes
        self.prediction_stat([name, px, end_time], pc)

        if len(np.unique(pc[:, 0]) > 1):
            pc = sort_by_length(pc)

            pc = pd.DataFrame(pc)
            pc.to_csv(
                name + "_instance.csv",
                header=["IDs", "X [A]", "Y [A]", "Z [A]"],
                index=False,
                sep=",",
            )

        self.upload_tasks += [
            asyncio.create_task(
                upload_to_aws(
                    self.aws_key,
                    self.aws_secret,
                    self.bucket,
                    folder_name,
                    name,
                    name + "_instance.csv",
                    remove_=False,
                )
            ),
            upload_to_aws(
                self.aws_key,
                self.aws_secret,
                self.bucket,
                "TARDIS_Prediction_Stats",
                None,
                "stat.json",
                remove_=False,
            ),
        ]


async def main():
    # AWS pre-setting
    aws_dir = "aws_czi.csv"
    aws = np.genfromtxt(aws_dir, delimiter=",", dtype=str)
    aws_key = aws[0]
    aws_secret = aws[1]
    client = Client()
    bucket = "cryoetportal-rawdatasets-dev"

    allocate_gpu = ["6", "7"]
    predict = "Membrane"
    start_from = 13143

    assert predict in ["Membrane", "Microtubule"]

    terminal = TardisLogo()
    terminal(title="Predict CIZ-Cryo-ET data-port datasets")
    process_tardis = ProcessTardisForCZI(
        aws=[aws_key, aws_secret, bucket], allocate_gpu=allocate_gpu, predict=predict
    )

    # Get all tomograms
    all_tomogram = list(
        Tomogram.find(
            client,
        )
    )

    # S3 URIs for MRCs
    urls = [t.s3_mrc_scale0 for t in all_tomogram]
    folder_names = [s[31:36] for s in urls]
    file_names = [s.split("/")[4] for s in urls]
    czi_px = [float(s.split("/")[6][12:]) for s in urls]

    terminal(
        text_1="Prediction:",
        text_3="Dataset: []",
        text_5=print_progress_bar(start_from, len(urls)),
    )

    tasks = []
    next_download = asyncio.create_task(get_from_aws(urls[start_from][31:]))
    for idx, (folder_name, file_name, px_czi) in enumerate(
        zip(folder_names, file_names, czi_px)
    ):
        if idx < start_from:
            continue

        terminal(
            text_1="Waiting for Data from AWS:",
            text_3=f"Dataset: [{file_name}]",
            text_6=print_progress_bar(idx, len(urls)),
        )
        await next_download

        terminal(
            text_1="Normalize image:",
            text_3=f"Dataset: [{file_name}]",
            text_6=print_progress_bar(idx, len(urls)),
        )
        data, px = load_image(file_name + ".mrc", False)
        data = process_tardis.pre_process(data, px, file_name)

        header = mrc_read_header(file_name + ".mrc")
        remove(file_name + ".mrc")

        if px != px_czi:
            px = px_czi

        # If there's a next file, start downloading it now
        if idx + 1 < len(urls):
            next_download = asyncio.create_task(get_from_aws(urls[idx + 1][31:]))

        start_time = time.time()
        data = asyncio.create_task(
            process_tardis.cnn_prediction(data, px, header, file_name, (idx, len(urls)))
        )

        tasks.append(
            asyncio.create_task(
                process_tardis.dist_prediction(
                    data, px, start_time, file_name, folder_name, (idx, len(urls))
                )
            )
        )

    await asyncio.gather(*tasks)


if __name__ == "__main__":
    asyncio.run(main())



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional

import torch
import torch.nn as nn

from tardis_em.cnn.model.convolution import (
    DoubleConvolution,
    RecurrentDoubleConvolution,
)
from tardis_em.cnn.model.init_weights import init_weights
from tardis_em.cnn.utils.utils import number_of_features_per_level


class DecoderBlockCNN(nn.Module):
    """
    CNN DECODER BUILDER FOR UNET

    Create a decoder block consisting of indicated number of deconvolutions followed
    by up-sampling and concatenation.

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        size (int): Size for the resampling.
        dropout (float, optional): Optional, dropout rate.
        components (str): String list of components from which convolution block
            is composed
        num_group (int): No. of groups for nn.GroupNorm()
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        conv_kernel: int,
        padding: int,
        size: int,
        dropout: Optional[float] = None,
        components="3gcr",
        num_group=8,
    ):
        super(DecoderBlockCNN, self).__init__()

        self.dropout = dropout

        """Build decoders"""
        if "3" in components:
            self.upscale = nn.Upsample(size=size, mode="trilinear", align_corners=False)
        elif "2" in components:
            self.upscale = nn.Upsample(size=size, mode="bilinear", align_corners=False)

        self.deconv_module = DoubleConvolution(
            in_ch=in_ch,
            out_ch=out_ch,
            block_type="decoder",
            kernel=conv_kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        """Optional Dropout"""
        if dropout is not None:
            self.dropout_layer = nn.Dropout(dropout)

        """Initialise the blocks"""
        for m in self.children():
            if m.__class__.__name__.find("Conv3d") != -1:
                continue
            if m.__class__.__name__.find("Conv2d") != -1:
                continue
            if m.__class__.__name__.find("GroupNorm") != -1:
                continue

            init_weights(m)

    def forward(self, encoder_features: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """
        Forward CNN decoder block for Unet

        Args:
            encoder_features (torch.Tensor): Residual connection.
            x (torch.Tensor): Image tensor before convolution.

        Returns:
            torch.Tensor: Image tensor after convolution.
        """

        x = encoder_features + self.deconv_module(self.upscale(x))

        if self.dropout is not None:
            x = self.dropout_layer(x)

        return x


class DecoderBlockRCNN(nn.Module):
    """
    RCNN DECODER BUILDER

    Create a decoder block consisting of indicated number of deconvolutions followed
    by upscale and connection with torch.cat().

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        size (int): Size for the resampling.
        dropout (float, optional): Optional, dropout rate.
        components (str): String list of components from which convolution block
            is composed
        num_group (int): No. of groups for nn.GroupNorm()
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        conv_kernel: int,
        padding: int,
        size: int,
        dropout: Optional[float] = None,
        components="3gcr",
        num_group=8,
    ):
        super(DecoderBlockRCNN, self).__init__()

        self.dropout = dropout

        """Build decoders"""
        if "3" in components:
            self.upscale = nn.Upsample(size=size, mode="trilinear", align_corners=False)
        elif "2" in components:
            self.upscale = nn.Upsample(size=size, mode="bilinear", align_corners=False)

        self.deconv_module = DoubleConvolution(
            in_ch=in_ch,
            out_ch=out_ch,
            block_type="decoder",
            kernel=conv_kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        self.deconv_res_module = RecurrentDoubleConvolution(
            in_ch=out_ch,
            out_ch=out_ch,
            block_type="decoder",
            kernel=conv_kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        """Optional Dropout"""
        if dropout is not None:
            self.dropout_layer = nn.Dropout(dropout)

        """Initialise the blocks"""
        for m in self.children():
            if m.__class__.__name__.find("Conv3d") != -1:
                continue
            if m.__class__.__name__.find("Conv2d") != -1:
                continue
            if m.__class__.__name__.find("GroupNorm") != -1:
                continue

            init_weights(m)

    def forward(self, encoder_features: torch.Tensor, x: torch.Tensor) -> torch.Tensor:
        """
        Forward RCNN decoder block

        Args:
            encoder_features (torch.Tensor): Residual connection.
            x (torch.Tensor): Image tensor before convolution.

        Returns:
            torch.Tensor: Image tensor after convolution.
        """
        x = self.upscale(x)
        x = self.deconv_module(x)

        x = encoder_features + x
        x = self.deconv_res_module(x)

        if self.dropout is not None:
            x = self.dropout_layer(x)

        return x


class DecoderBlockUnet3Plus(nn.Module):
    """
    CNN DECODER BUILDER FOR UNET3PLUS

    Create a decoder block consisting of the indicated number of deconvolutions followed
    by upscale and connection with torch.cat().

    Args:
        in_ch (int): Number of input channels.
        out_ch (int): Number of output channels.
        size (int): Size for the resampling.
        dropout (float, optional): Optional, dropout rate.
        components (str): String list of components from which convolution block
            is composed.
        num_group (int): Number of groups inc CNN block.
        num_layer (int): Number of CNN layers.
        encoder_feature_ch (list): List of encode outputs.
        num_group (int): No. of groups for nn.GroupNorm()
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        conv_kernel: int,
        padding: int,
        size: int,
        components: str,
        num_group: int,
        num_layer: int,
        encoder_feature_ch: list,
        attn_features=False,
        dropout: Optional[float] = None,
    ):
        super(DecoderBlockUnet3Plus, self).__init__()

        self.attn_features = attn_features
        self.dropout = dropout

        """Main Block Up-Convolution"""
        if "3" in components:
            self.upscale = nn.Upsample(size=size, mode="trilinear", align_corners=False)
        elif "2" in components:
            self.upscale = nn.Upsample(size=size, mode="bilinear", align_corners=False)

        self.deconv = DoubleConvolution(
            in_ch=in_ch,
            out_ch=out_ch,
            block_type="decoder",
            kernel=conv_kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        """Skip-Connection Encoders"""
        num_layer = num_layer - 1
        pool_kernels = [None, 2, 4, 8, 16, 32]
        pool_kernels = pool_kernels[:num_layer]

        self.encoder_max_pool = nn.ModuleList([])
        self.encoder_feature_conv = nn.ModuleList([])

        for i, en_in_channel in enumerate(encoder_feature_ch):
            pool_kernel = pool_kernels[i]

            if pool_kernel is not None and "3" in components:
                max_pool = nn.MaxPool3d(
                    kernel_size=pool_kernel, stride=pool_kernel, ceil_mode=True
                )
            elif pool_kernel is not None and "2" in components:
                max_pool = nn.MaxPool2d(
                    kernel_size=pool_kernel, stride=pool_kernel, ceil_mode=True
                )
            else:
                max_pool = None

            conv = DoubleConvolution(
                in_ch=en_in_channel,
                out_ch=out_ch,
                block_type="decoder",
                kernel=conv_kernel,
                padding=padding,
                components=components,
                num_group=num_group,
            )
            self.encoder_max_pool.append(max_pool)
            self.encoder_feature_conv.append(conv)

        if attn_features is not None:
            self.attn_conv = DoubleConvolution(
                in_ch=out_ch + in_ch,
                out_ch=out_ch,
                block_type="decoder",
                kernel=conv_kernel,
                padding=padding,
                components=components,
                num_group=num_group,
            )

        """Optional Dropout"""
        if dropout is not None:
            self.dropout_layer = nn.Dropout(dropout)

        """Initialise the blocks"""
        for m in self.children():
            if m.__class__.__name__.find("Conv3d") != -1:
                continue
            if m.__class__.__name__.find("Conv2d") != -1:
                continue
            if m.__class__.__name__.find("GroupNorm") != -1:
                continue

            init_weights(m)

    def forward(self, x: torch.Tensor, encoder_features: list) -> torch.Tensor:
        """
        Forward CNN decoder block for Unet3Plus

        Args:
            x (torch.Tensor): Image tensor before convolution.
            encoder_features (list): Residual connection from the encoder.

        Returns:
            torch.Tensor: Image tensor after convolution.
        """

        """Main Block"""
        if self.attn_features:
            x_attn = self.upscale(x)
            x = self.deconv(x_attn)
        else:
            x = self.deconv(self.upscale(x))

        """Skip-Connections Encoder"""
        for i, encoder in enumerate(encoder_features):
            if self.encoder_max_pool[i] is not None:
                encoder = self.encoder_feature_conv[i](
                    self.encoder_max_pool[i](encoder)
                )
            else:
                encoder = self.encoder_feature_conv[i](encoder)
            x = x + encoder

        if self.attn_features:
            x = self.attn_conv(torch.cat((x, x_attn), dim=1))

        # Additional Dropout
        if self.dropout is not None:
            x = self.dropout_layer(x)

        return x


def build_decoder(
    conv_layers: int,
    conv_layer_scaler: int,
    components: str,
    num_group: int,
    conv_kernel: int,
    padding: int,
    sizes: list,
    dropout: Optional[float] = None,
    deconv_module="CNN",
    attn_features=False,
):
    """
    Decoder wrapper for the entire CNN model.

    Create a decoder block from a number of convolution and convolution modules.
    The Decoder block is followed by upscale(interpolation) and joining with
    torch.cat()

    Args:
        conv_layers (int): Number of deconvolution layers.
        conv_layer_scaler (int): Number of channels by which each CNN block is scaled up.
        components (str): Components that are used for deconvolution block.
        conv_kernel (int): Convolution kernel size.
        padding (int): Padding size for the convolution.
        sizes (list): List of tensor sizes for upscale.
        num_group (int): Num. of groups for the nn.GroupNorm.
            None -> if nn.GroupNorm is not used.
        dropout (float, Optional): Dropout value.
        deconv_module: Module of the deconvolution for the decoder.
        attn_features:

    Returns:
        nn.ModuleList: List of decoder blocks.
    """
    decoders = []
    feature_map = number_of_features_per_level(
        channel_scaler=conv_layer_scaler, num_levels=conv_layers
    )
    feature_map = list(reversed(feature_map))

    """Forward decoder"""
    if deconv_module == "CNN":
        # Unet decoder
        for i in range(len(feature_map) - 1):
            in_ch = feature_map[i]
            out_ch = feature_map[i + 1]
            size = sizes[i]

            decoder = DecoderBlockCNN(
                in_ch=in_ch,
                out_ch=out_ch,
                conv_kernel=conv_kernel,
                padding=padding,
                size=size,
                dropout=dropout,
                components=components,
                num_group=num_group,
            )
            decoders.append(decoder)
    elif deconv_module == "RCNN":
        # ResNet decoder
        for i in range(len(feature_map) - 1):
            in_ch = feature_map[i]
            out_ch = feature_map[i + 1]
            size = sizes[i]

            decoder = DecoderBlockRCNN(
                in_ch=in_ch,
                out_ch=out_ch,
                conv_kernel=conv_kernel,
                padding=padding,
                size=size,
                dropout=dropout,
                components=components,
                num_group=num_group,
            )

            decoders.append(decoder)
    elif deconv_module == "unet3plus":
        # Unet3Plus decoder
        idx_en = 1

        for i in range(len(feature_map) - 1):
            # Main Module features
            in_ch = feature_map[i]
            out_ch = feature_map[i + 1]
            size = sizes[i]

            # Encoder De Convolution
            encoder_feature_ch = feature_map[idx_en:]
            idx_en += 1

            decoder = DecoderBlockUnet3Plus(
                in_ch=in_ch,
                out_ch=out_ch,
                conv_kernel=conv_kernel,
                padding=padding,
                size=size,
                components=components,
                num_group=num_group,
                num_layer=conv_layers,
                encoder_feature_ch=encoder_feature_ch,
                attn_features=attn_features,
                dropout=dropout,
            )
            decoders.append(decoder)

    return nn.ModuleList(decoders)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from typing import Optional, Tuple, Union

import numpy as np
from scipy.interpolate import splev, splprep
from scipy.spatial.distance import cdist
from sklearn.neighbors import KDTree

from tardis_em.dist_pytorch.utils.utils import pc_median_dist
from tardis_em.utils.errors import TardisError


def resample_filament(points, spacing_size):
    """
    Resamples points for each filament so they have the given spacing size.

    Parameters:
    points (np.array): Array of shape [N, 4] where each column is [ID, X, Y, Z].
    spacing_size (float): Desired spacing size between points.

    Returns:
    np.array: Resampled array with the same structure.
    """
    unique_ids = np.unique(points[:, 0])
    resampled_points = []

    for filament_id in unique_ids:
        filament_points = points[points[:, 0] == filament_id][:, 1:]
        if filament_points.shape[0] < 2:
            resampled_points.append(points[points[:, 0] == filament_id])
            continue

        cumulative_distances = np.cumsum(np.sqrt(np.sum(np.diff(filament_points, axis=0) ** 2, axis=1)))
        cumulative_distances = np.insert(cumulative_distances, 0, 0)

        num_new_points = int(cumulative_distances[-1] // spacing_size)
        new_points = [filament_points[0]]

        for i in range(1, num_new_points + 1):
            target_distance = i * spacing_size
            idx = np.searchsorted(cumulative_distances, target_distance)
            if idx >= len(cumulative_distances):
                continue

            if cumulative_distances[idx] == target_distance:
                new_points.append(filament_points[idx])
            else:
                ratio = (target_distance - cumulative_distances[idx - 1]) / (
                            cumulative_distances[idx] - cumulative_distances[idx - 1])
                new_point = filament_points[idx - 1] + ratio * (filament_points[idx] - filament_points[idx - 1])
                new_points.append(new_point)

        new_points = np.array(new_points)
        filament_ids = np.full((new_points.shape[0], 1), filament_id)
        resampled_points.append(np.hstack((filament_ids, new_points)))

    return np.vstack(resampled_points)


class FilterConnectedNearSegments:
    """
    Connect splines based on spline trajectory and splines end distances.
    """

    def __init__(self, distance_th=1000, cylinder_radius=150):
        """
        Initialize the class with the distance threshold and cylinder radius parameters.

        Args:
            distance_th (int): Maximum distance between two spline endpoints for them
            to be considered for merging.
            cylinder_radius (int): Maximum distance between spline endpoint and the line
            connecting its neighboring endpoints for them to be considered for merging.
        """
        self.distance_th = distance_th
        self.cylinder_radius = cylinder_radius

    @staticmethod
    def _in_cylinder(point: np.ndarray, axis: tuple, r: int, h: int) -> bool:
        """
        Fast check if point is inside a cylinder by projecting point on cylinder
        volume in 3D.

        Args:
            point (np.ndarray): Point in 3D to project on cylinder.
            axis (tuple): Cylinder axis of orientation.
            r (int): Cylinder radius.
            h (int): Cylinder length.

        Returns:
            bool: If True, given point can be found in a cylinder.
        """
        # project the point onto the axis
        ax1 = (point - axis[0]) + 1e-16
        ax2 = (axis[1] - axis[0]) + 1e-16

        d = np.dot(ax1, ax2) / np.linalg.norm(ax2)

        # check if d is within the range [0, h]
        if d < 0 or d > h:
            return False

        # calculate the perpendicular distance
        d = np.linalg.norm(ax1 - d * (ax2) / np.linalg.norm(ax2))

        # check if d is less than the radius
        return d <= r

    @staticmethod
    def _remove_duplicates(d: dict) -> dict:
        """
        Remove duplicate splines

        Args:
            d (dict): Dictionary containing splines.

        Returns:
            dict: dictionary containing unique splines.
        """
        new_d = {}
        [
            new_d.update({k: v.tolist()})
            for k, v in d.items()
            if v.tolist() not in new_d.values()
        ]

        return new_d

    def splines_direction(self, spline1: list, spline2: list) -> bool:
        """
        Check if two splines facing the same direction. Knowing that spline1 facing
        a direction of spline2, check if spline2 also face a direction of spline1.

        Args:
            spline1 (list): Sorted array of point with 3D coordinates.
            spline2 (list): Sorted array of point with 3D coordinates.

        Returns:
            bool: If true, given splines facing the same direction
        """
        # Check 01 - 01 & Check 01 - 10
        ax = [
            (np.array(spline2[1]), np.array(spline2[0])),
            (np.array(spline2[1]), np.array(spline2[0])),
        ]
        points = [np.array(spline1[0]), np.array(spline1[-1])]
        s201_s101, s201_s110 = [
            self._in_cylinder(
                point=p, axis=a, r=self.cylinder_radius, h=self.distance_th
            )
            for p, a in zip(points, ax)
        ]

        # Check 10 - 01 and Check 10 - 10
        ax = [
            (np.array(spline2[-2]), np.array(spline2[-1])),
            (np.array(spline2[-2]), np.array(spline2[-1])),
        ]
        points = [np.array(spline1[0]), np.array(spline1[-1])]
        s210_s101, s210_s110 = [
            self._in_cylinder(
                point=p, axis=a, r=self.cylinder_radius, h=self.distance_th
            )
            for p, a in zip(points, ax)
        ]

        # Check if splines facing same direction in any way
        return np.any((s201_s101, s201_s110, s210_s101, s210_s110))

    def marge_splines(
        self, point_cloud: np.ndarray, omit_border: int, initial: bool
    ) -> np.ndarray:
        """
        Connect splines in the point_cloud that are close enough to each other
         and return the reordered splines.

        Example:
            - Get dictionary with {id: [sorted list of points]}

            - While there is more than 1 spline in the dict:
                - Pick first spline in a dict.
                - Select all ends
                - Calc. end distance to all other ends.

                - If any ends within threshold distance:
                    - Calculate initial end vector
                    - Calculate distance to the cylinder

                    - If any picked ends are within cylinder radius
                        - Pick one with the smallest distance
                        - Save two spline IDs to connected

                - If Connected ID list not empty:
                    - Marge and sort points from two IDs
                    - Check tortuosity

                    - If tortuosity <= 1.5:
                        - Add to the connected dict
                        - Remove from dict
                        - start over

                - If connected ID list is empty or tortuosity > 1.5:
                    - Add only initial spline to the connected dict
                    - Remove from dict
                    - start over

        Args:
            point_cloud (np.ndarray): Array with segmented and sorted point cloud
                of a shape [ID, X, Y, Z].
            omit_border (int): In A, distance from the border as a limit not to
                connect splines.
            initial (bool): Initial run for the operation.

        Returns:
            np.ndarray: Array of segmented point cloud with connected splines
            that fit the criteria
        """
        # Find Z bordered to filter out MT connection at the borders
        MIN_Z, MAX_Z = np.min(point_cloud[:, 3]), np.max(point_cloud[:, 3])

        # Create a dictionary to store spline information
        # Iterate through the point cloud and add points to their respective splines
        if initial:
            splines_list_df = {}
            for point in point_cloud:
                id_, x, y, z = point
                if id_ not in splines_list_df:
                    splines_list_df[id_] = []
                splines_list_df[id_].append([x, y, z])

            splines_list = {}
            for i in splines_list_df:
                value = splines_list_df[i]
                if len(value) > 5:
                    splines_list[i] = value
        else:
            splines_list = {}
            for point in point_cloud:
                id_, x, y, z = point
                if id_ not in splines_list:
                    splines_list[id_] = []
                splines_list[id_].append([x, y, z])

        # Iterate throw every spline in the list
        merge_splines = {}
        spline_id = 0
        while len(splines_list) > 1:
            key = list(splines_list.keys())[0]
            value = splines_list[key]  # Pick first spline in the dictionary
            end01 = value[0]
            end10 = value[-1]

            end01_list = [list(x)[1][0] for x in splines_list.items()]
            end10_list = [list(x)[1][-1] for x in splines_list.items()]

            # Check if any ends is within threshold distance
            end01_list01 = np.sqrt(
                np.sum((np.asarray(end01_list) - np.asarray(end01)) ** 2, axis=1)
            )
            end01_list01 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end01_list01)
                if dist <= 1000 and id_ != key
            ]
            end01_list10 = np.sqrt(
                np.sum((np.asarray(end10_list) - np.asarray(end01)) ** 2, axis=1)
            )
            end01_list10 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end01_list10)
                if dist <= self.distance_th and id_ != key
            ]
            end10_list01 = np.sqrt(
                np.sum((np.asarray(end01_list) - np.asarray(end10)) ** 2, axis=1)
            )
            end10_list01 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end10_list01)
                if dist <= self.distance_th and id_ != key
            ]
            end10_list10 = np.sqrt(
                np.sum((np.asarray(end10_list) - np.asarray(end10)) ** 2, axis=1)
            )
            end10_list10 = [
                {id_: dist}
                for id_, dist in zip(list(splines_list.keys()), end10_list10)
                if dist <= self.distance_th and id_ != key
            ]

            # Check if any of the point is within the cylinder and get the closest one
            splines_to_merge = []
            end_lists = [end01_list01, end01_list10, end10_list01, end10_list10]

            for end_list in end_lists:
                for i in end_list:
                    m_id = list(i.keys())[
                        -1 if end_list in [end10_list01, end10_list10] else 0
                    ]
                    m_end = list(splines_list[m_id])

                    not_at_the_border = np.all(
                        [
                            (
                                m_end[
                                    (
                                        -1
                                        if end_list in [end10_list01, end10_list10]
                                        else 0
                                    )
                                ][2]
                                - MIN_Z
                            )
                            >= omit_border,
                            (
                                MAX_Z
                                - m_end[
                                    (
                                        -1
                                        if end_list in [end10_list01, end10_list10]
                                        else 0
                                    )
                                ][2]
                            )
                            >= omit_border,
                        ]
                    )

                    if not_at_the_border:
                        points = np.array(
                            m_end[-1 if end_list in [end10_list01, end10_list10] else 0]
                        )
                        axis = (
                            np.array(
                                value[
                                    (
                                        -2
                                        if end_list in [end10_list01, end10_list10]
                                        else 1
                                    )
                                ]
                            ),
                            np.array(
                                value[
                                    (
                                        -1
                                        if end_list in [end10_list01, end10_list10]
                                        else 0
                                    )
                                ]
                            ),
                        )

                        in_cylinder = self._in_cylinder(
                            point=points,
                            axis=axis,
                            r=self.cylinder_radius,
                            h=self.distance_th,
                        )
                        if in_cylinder:
                            splines_to_merge.append([m_id, m_end])

            # Check if any ends fit criteria for connection
            if len(splines_to_merge) == 1:  # Merge
                # Check if selected spline facing the same direction
                same_direction = self.splines_direction(value, splines_to_merge[0][1])

                # Connect splines
                if same_direction:
                    merged_spline = np.concatenate((value, splines_to_merge[0][1]))
                    merge_splines[spline_id] = sort_segment(merged_spline)

                    del splines_list[splines_to_merge[0][0]]
            elif len(splines_to_merge) > 1:  # If more than one find best
                if len(np.unique([x[0] for x in splines_to_merge])) == 1:
                    # Check if selected spline facing the same direction
                    same_direction = self.splines_direction(
                        value, splines_to_merge[0][1]
                    )

                    # Connect splines
                    if same_direction:
                        merged_spline = np.concatenate((value, splines_to_merge[0][1]))

                        merge_splines[spline_id] = sort_segment(merged_spline)
                        del splines_list[splines_to_merge[0][0]]
                else:
                    end_lists = {}
                    for d in np.concatenate(
                        [end01_list01, end01_list10, end10_list01, end10_list10]
                    ):
                        end_lists.update(d)

                    # Check which splines facing the same direction
                    same_direction = []
                    for d in end_lists:
                        same_direction.append(
                            self.splines_direction(value, splines_list[d])
                        )

                    # Pick splines with the smallest distance that facing same direction
                    end_lists_id = [x for x, b in zip(end_lists, same_direction) if b]
                    same_direction = np.any(same_direction)

                    if len(end_lists_id) > 0:
                        end_lists_id = min(end_lists_id, key=lambda x: end_lists[x])
                        merged_spline = np.concatenate(
                            (value, splines_list[end_lists_id])
                        )

                        merge_splines[spline_id] = sort_segment(merged_spline)
                        del splines_list[end_lists_id]
            else:  # No merge found
                same_direction = True
                merge_splines[spline_id] = sort_segment(value)

            if not same_direction:
                merge_splines[spline_id] = sort_segment(value)

            del splines_list[key]
            spline_id += 1

        # Add last spline to the new list
        try:
            key = list(splines_list.keys())[0]
            merge_splines[spline_id] = splines_list[key]
        except IndexError:
            pass

        return np.concatenate(
            [
                np.hstack((np.repeat(id_, len(array)).reshape(-1, 1), array))
                for id_, array in merge_splines.items()
            ]
        )

    def __call__(self, point_cloud: np.ndarray, omit_border=0):
        past_l = 0
        while len(np.unique(point_cloud[:, 0])) != past_l:
            if past_l == 0:
                past_l = len(np.unique(point_cloud[:, 0]))
                point_cloud = self.marge_splines(
                    point_cloud=point_cloud, omit_border=omit_border, initial=True
                )
            else:
                past_l = len(np.unique(point_cloud[:, 0]))
                point_cloud = self.marge_splines(
                    point_cloud=point_cloud, omit_border=omit_border, initial=False
                )
        return point_cloud


class FilterSpatialGraph:
    """
    Calculate length of each spline and distance between all splines ends.

    This clas iterate over all splines in array [ID, X, Y, Z] by given ID.
    Firstly if specified during initialization, class calculate distance
    between all splines ends and use it to define which splines are obviously
    broken.
    Then it calculate length of all the splines (also new one) and filter out
    ones that are too short.
    """

    def __init__(
        self,
        connect_seg_if_closer_then=1000,
        cylinder_radius=200,
        filter_short_segments=1000,
    ):
        self.connect_seg_if_closer_then = connect_seg_if_closer_then
        self.filter_short_segments = filter_short_segments

        self.marge_splines = FilterConnectedNearSegments(
            distance_th=connect_seg_if_closer_then, cylinder_radius=cylinder_radius
        )

    def __call__(self, segments: np.ndarray) -> np.ndarray:
        """
        Connect splines that have their end's in close distance and remove
        splines that are too short.

        Args:
            segments (np.ndarray): Array of points with ID label of shape [ID, X, Y, Z]

        Returns:
            np.ndarray: Filtered array of connected MTs
        """
        """Do iterative optimization split 150 degree connection / marge"""
        # Split 150 degree connections
        loop_ = True
        while loop_:
            loop_, segments = cut_150_degree(segments)

        # Connect segments with ends close to each other
        border = [np.min(segments[:, 3]), np.max(segments[:, 3])]
        border = (border[1] - border[0]) / 50

        if self.connect_seg_if_closer_then > 0:
            segments = self.marge_splines(point_cloud=segments, omit_border=border)
            segments = reorder_segments_id(segments)

        """Remove too short splines"""
        if self.filter_short_segments > 0:
            length = []
            for i in np.unique(segments[:, 0]):
                length.append(
                    total_length(segments[np.where(segments[:, 0] == int(i))[0], 1:])
                )

            length = [
                id_ for id_, i in enumerate(length) if i > self.filter_short_segments
            ]

            new_seg = []
            for i in length:
                new_seg.append(segments[np.where(segments[:, 0] == i), :])

            if len(new_seg) > 0:
                segments = np.hstack(new_seg)[0, :]
                segments = reorder_segments_id(segments)

        # Split 150 degree connections
        loop_ = True
        while loop_:
            loop_, segments = cut_150_degree(segments)

        return reorder_segments_id(segments)


class SpatialGraphCompare:
    """
    Compare two spatial graphs and output filtered-out array's of splines based
    on similarity.

    This class take as na input two arrays of shape [n, 3 or 4] for 2D or 3D
    point cloud. This arrays contain [ID x X x Y x Z] dimensions.

    The comparison is archived by calculating cdist for all splines from one spatial
    graph to all splines from second spatial graph. And for each spline it output
    probability of similarity and splines id's.

    The Probability is calculated as a ration of points (in threshold contact)
    to all points in spline.

    The selection threshold for the spline interaction is given between 0 and 1.
    """

    def __init__(self, distance_threshold: int, interaction_threshold: float):
        self.dist_th = distance_threshold
        self.inter_th = interaction_threshold

    def _compare_spatial_graphs(
        self, spatial_graph_1: np.ndarray, spatial_graph_2: np.ndarray
    ) -> list:
        """
        Wrapper to compare all MT's between two spatial graphs

        Args:
            spatial_graph_1 (np.ndarray): Spatial graph 1.
            spatial_graph_2 (np.ndarray): Spatial graph 2.

        Returns:
            list: List of MT from spatial graph 1 that match spatial graph 2.
        """
        match_sg1_sg2 = []

        for i in np.unique(spatial_graph_1[:, 0]):
            sg1_spline = spatial_graph_1[spatial_graph_1[:, 0] == i, :]
            iou = []

            for j in np.unique(spatial_graph_2[:, 0]):
                sg2_spline = spatial_graph_2[spatial_graph_2[:, 0] == j, :]
                iou.append(
                    compare_splines_probability(
                        sg1_spline[:, 1:], sg2_spline[:, 1:], self.dist_th
                    )
                )

            ids = [
                id for id, i in enumerate(iou) if np.sum(i) > 0 and i >= self.inter_th
            ]
            match_sg1_sg2.append([i, ids])

        return match_sg1_sg2

    def __call__(
        self, amira_sg: np.ndarray, tardis_sg: np.ndarray
    ) -> Tuple[list, list]:
        """
        Compute comparison of Amira and Tardis spatial graphs and output tuple of
        arrays with different selected MTs:
            - Label1: MT taken from the Tardis (matches Amira)
            - Label2: MT taken from the Amira (matches Tardis)
            - Label3: MT in Tardis without match
            - Label4: MT in Amira without match

        Args:
            amira_sg (np.ndarray): Spatial graph [ID, X, Y, Z] from Amira.
            tardis_sg (np.ndarray): Spatial graph [ID, X, Y, Z] from tardis_em.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: Tuple of all arrays
        """
        label = [
            "TardisFilterBasedOnAmira",
            "TardisNoise",
            "AmiraFilterBasedOnTardis",
            "AmiraNoise",
        ]

        """Amira splines scores"""
        amira_comp = self._compare_spatial_graphs(
            spatial_graph_1=amira_sg, spatial_graph_2=tardis_sg
        )

        """Compare Tardis with Amira"""
        tardis_comp = self._compare_spatial_graphs(
            spatial_graph_1=tardis_sg, spatial_graph_2=amira_sg
        )

        # Select all splines from Tardis that match Amira
        match = [x[0] for x in tardis_comp if x[1] != []]
        tardis_match = np.stack([x for x in tardis_sg if x[0] in match])

        noise = [x[0] for x in tardis_comp if x[1] == []]
        tardis_noise = np.stack([x for x in tardis_sg if x[0] in noise])

        # Select all splines from Amira that match Tardis
        match = [x[0] for x in amira_comp if x[1] != []]
        amira_match = np.stack([x for x in amira_sg if x[0] in match])

        noise = [x[0] for x in amira_comp if x[1] == []]
        amira_noise = np.stack([x for x in amira_sg if x[0] in noise])

        spatial_graphs = [
            g
            for g in (tardis_match, tardis_noise, amira_match, amira_noise)
            if g is not None
        ]
        label = [
            l
            for g, l in zip(
                (tardis_match, tardis_noise, amira_match, amira_noise), label
            )
            if g is not None
        ]

        return spatial_graphs, label


def compare_splines_probability(
    spline_1: np.ndarray, spline_2: np.ndarray, threshold=100
):
    """
    Compare two splines and calculate probability of how likely given two
    splines are the same line given array of points for same or similar splines
    with no matching coordinates of points.

    Calculates the probability of two splines being similar by comparing
    the distance between their points and taking the mean of the matching
    points below a threshold.

    Parameters:
        spline_1 (np.ndarray): The first spline to compare, represented
            as an array of points.
        spline_2 (np.ndarray): The second spline to compare, represented
            as an array of points.
        threshold (int): The maximum distance between points for them to be
            considered matching.

    Returns:
        float: The probability of the splines being similar, ranging from 0.0 to 1.0.
    """
    if len(spline_1) == 0 or len(spline_2) == 0:
        return 0.0

    # Calculating distance matrix between points of 2 splines
    dist_matrix = cdist(spline_1, spline_2)

    # Check how many points on spline1 match spline2
    m_s1 = [1 if x < threshold else 0 for x in np.min(dist_matrix, axis=1)]

    # If no matching points probability is 0
    if sum(m_s1) == 0:
        return 0.0

    # Calculating intersection using % of the matching point below a threshold
    probability = sum(m_s1) / len(spline_1)

    return probability


def smooth_spline(points: np.ndarray, s=0.5):
    """
    Spline smoothing given an 's' smoothness factor.

    Args:
        points (np.ndarray): Point array [(ID), X, Y, Z] with optional ID and Z
        dimension.
        s (float): Smoothness factor.

    Returns:
        Returns: Smooth spline
    """
    if points.shape[1] == 4:  # [ID, X, Y, Z]
        id_ = int(points[0, 0])
        points = points[:, 1:]
        norm_pc = pc_median_dist(points)
        points = points / norm_pc

        t_before = tortuosity(points)
        try:
            tck, u = splprep(points.T, s=s)
            spline = np.stack(splev(u, tck)).T
        except ValueError:
            spline = points

        spline = spline * norm_pc
        t_after = tortuosity(spline)
        ids = np.zeros((len(spline), 1))
        ids[:, 0] = id_

        # Sanity check if spline smoothing failed
        if t_after > t_before:
            return np.hstack((ids, points * norm_pc))
        return np.hstack((ids, spline))
    else:  # [X, Y, Z]
        tck, u = splprep(points.T)

        return np.stack(splev(u, tck)).T


def sort_segment(coord: np.ndarray) -> np.ndarray:
    """
    Sorting of the point cloud based on number of connections followed by
    searching of the closest point with cdist.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        np.ndarray: Array of point in line order.
    """
    if len(coord) < 2:
        return coord

    new_c = []
    for i in range(len(coord) - 1):
        if i == 0:
            id = np.where(
                [sum(i) for i in cdist(coord, coord)]
                == max([sum(i) for i in cdist(coord, coord)])
            )[0]

            new_c.append(coord[id[0]])
            coord = np.delete(coord, id[0], 0)

        kd = KDTree(coord)
        points = kd.query(np.expand_dims(new_c[len(new_c) - 1], 0))[1][0][0]

        new_c.append(coord[points])
        coord = np.delete(coord, points, 0)

    if len(new_c) > 0:
        return np.stack(new_c)
    else:
        return coord


def reorder_segments_id(
    coord: np.ndarray,
    order_range: Optional[list] = None,
    order_list: Optional[Union[list, np.ndarray]] = None,
) -> np.ndarray:
    """
    Reorder list of segments to remove missing IDs

    E.g. Change IDs from [1, 2, 3, 5, 6, 8] to [1, 2, 3, 4, 5, 6]

    Args:
        coord: Array of points in 3D or 3D with their segment ID
        order_range: Costume id range for reordering
        order_list: List of reorder IDs to match to coord.

    Returns:
        np.ndarray: Array of points with reordered IDs values
    """
    df = np.unique(coord[:, 0])

    if order_range is None:
        df_range = np.asarray(range(0, len(df)), dtype=df.dtype)
    else:
        df_range = np.asarray(range(order_range[0], order_range[1]), dtype=df.dtype)

    if order_list is None:
        for id, i in enumerate(coord[:, 0]):
            coord[id, 0] = df_range[np.where(df == i)[0][0]]
    else:
        ordered_coord = []
        for i, new_id in zip(df, order_list):
            line = coord[np.where(coord[:, 0] == i)[0], :]
            line[:, 0] = new_id
            ordered_coord.append(line)
        coord = np.concatenate(ordered_coord)
        coord = reorder_segments_id(coord)

    return coord


def tortuosity(coord: np.ndarray) -> float:
    """
    Calculate spline tortuosity.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        float: Spline curvature measured with tortuosity.
    """
    if len(coord) <= 1:
        return 1.0

    length = total_length(coord) + 1e-16
    end_length = np.sqrt(np.sum((coord[0] - coord[-1]) ** 2) + 1e-16)

    return length / end_length


def total_length(coord: np.ndarray) -> float:
    """
    Calculate total length of the spline.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        float: Spline length.
    """
    length = 0
    c_len = len(coord) - 1

    for id, _ in enumerate(coord):
        if id == c_len:
            break

        # sqrt((x2 - x1)2 + (y2 - y1)2 + (z2 - z1)2)
        length += np.sqrt(np.sum((coord[id] - coord[id + 1]) ** 2))

    return length


def length_list(coord: np.ndarray) -> list:
    """
    Calculate total length of all splines and return it as a list.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        list: Spline length list.
    """
    spline_length_list = []

    for i in np.unique(coord[:, 0]):
        points = coord[np.where(coord[:, 0] == i)[0], 1:]
        spline_length_list.append(total_length(points))

    return spline_length_list


def angle_between_vectors(v1, v2):
    """
    Calculate the angle in degrees between two vectors.

    This function uses the dot product and the magnitudes of the vectors
    to calculate the angle between them according to the formula:

        cos(theta) = (A . B) / (||A|| ||B||)

    Args:
        v1(np.ndarray): First input vector.
        v2(np.ndarray): Second input vector.

    Returns:
        float The angle in degrees between vector 'v1' and 'v2'.
    """
    # Calculate the dot product of vectors v1 and v2
    dot_product = np.dot(v1, v2)

    # Calculate the magnitude (norm) of vectors
    magnitude_v1 = np.linalg.norm(v1) + 1e-16
    magnitude_v2 = np.linalg.norm(v2) + 1e-16

    # Calculate angle
    cos_theta = dot_product / (magnitude_v1 * magnitude_v2)
    angle = np.arccos(np.clip(cos_theta, -1.0, 1.0))

    return np.degrees(angle)


def cut_150_degree(segments_array: np.ndarray):
    """
    Cut segments based on angles between adjacent vectors.

    Given an array of line segments, this function calculates angles between
    adjacent vectors in each line. If the angle is less than or equal to 150
    degrees, the segment is cut into two new segments.

    Args:
    segments_array(np. ndarray): Array of line segments where the first column
    indicates the segment id and the remaining columns represent
    the coordinates of points.

    Args:
        segments_array:

    Returns:
        Tuple[bool, np.ndarray]: Indicates whether any segment was cut,
            and New array of cut segments.
    """

    cut_segments = []
    loop_ = False

    # Loop through unique segment IDs
    for i in np.unique(segments_array[:, 0]):
        pc_ = segments_array[np.where(segments_array[:, 0] == i)[0], 1:]

        angles_ = [180]

        # Calculate angles for each line segment
        for j in range(len(pc_) - 2):
            angles_.append(
                angle_between_vectors(
                    np.array(pc_[j]) - np.array(pc_[j + 1]),
                    np.array(pc_[j + 2]) - np.array(pc_[j + 1]),
                )
            )
        angles_.append(180)

        # Check if any angle is less than or equal to 150 degrees
        if len([id_ for id_, k in enumerate(angles_) if k <= 150]) > 0:
            loop_ = True

            # Find the minimum angle and cut the segment
            min_angle_idx = np.where(angles_ == np.min(angles_))[0][0]
            cut_segments.append(pc_[: min_angle_idx + 1, :])
            cut_segments.append(pc_[min_angle_idx + 1 :, :])
        else:
            cut_segments.append(pc_)

    # Filter out single-point segments
    cut_segments = [c for c in cut_segments if len(c) > 1]

    # Create the output array
    return loop_, np.vstack(
        [
            np.hstack((np.repeat(id_, len(c)).reshape(-1, 1), c))
            for id_, c in enumerate(cut_segments)
        ]
    )


def sort_by_length(coord):
    """
    Sort all splines by their length.

    Args:
        coord: Array of coordinates.

    Returns:
        np.ndarray: sorted and reorder splines.
    """
    length_list = []
    for i in np.unique(coord[:, 0]):
        length_list.append(total_length(coord[np.where(coord[:, 0] == i)[0], 1:]))

    sorted_id = np.argsort(length_list)

    sorted_list = [coord[np.where(coord[:, 0] == i)[0], 1:] for i in sorted_id]
    sorted_list = [
        np.hstack((np.repeat(i, len(sorted_list[i])).reshape(-1, 1), sorted_list[i]))
        for i in range(len(sorted_list))
    ]

    return np.concatenate(sorted_list)


class ComputeConfidenceScore:
    @staticmethod
    def _angle_smoothness(tangents):
        """
        Calculate the smoothness of a sequence of vectors based on the angles
        between consecutive vectors.

        The smoothness is computed as 1 minus the standard deviation of the angles between
        consecutive vectors. The angles are calculated using the dot product and
        magnitude of the vectors. A higher smoothness value indicates a more
        consistent alignment between consecutive vectors, whereas a lower smoothness value
        indicates more variability in the alignment.

        Args:
            tangents (np.ndarray): A 2D numpy array of shape (n, m), where n is
            the number of vectors and m is the dimension of each vector.
            Each row in the array represents a vector.

        Returns:
            float: A scalar value representing the smoothness of the sequence of vectors.
                Ranges from 0 to 1, with higher values indicating a smoother sequence.
        """
        if (
            not isinstance(tangents, np.ndarray)
            or tangents.ndim != 2
            or tangents.shape[0] < 2
        ):
            return 1.0
        else:
            magnitudes = np.linalg.norm(tangents, axis=1)
            non_zero_vectors = tangents[magnitudes > 0]

            # All vectors are approximately zero or there's only one non-zero vector
            if non_zero_vectors.shape[0] < 2:
                return 1.0

            angles = np.arccos(
                np.einsum("ij,ij->i", tangents[:-1], tangents[1:])
                / (
                    np.linalg.norm(tangents[:-1], axis=1)
                    * np.linalg.norm(tangents[1:], axis=1)
                )
            )
            smoothness = 1 - np.std(angles)
            return smoothness

    @staticmethod
    def normalized_length(points: np.ndarray, min_l: float, max_l: float):
        """
        Calculate and normalize the total length of a sequence of points.
        The total length is calculated using the `total_length` function.
        The result is then normalized to a value between 0 and 1 based on the provided
        minimum and maximum length values.

        Args:
            points (np.ndarray):
            min_l (float): The minimum length value.
            max_l (float): The maximum length value.
        """
        length = total_length(points)
        return (length - min_l) / (max_l - min_l + 1e-16)

    def combined_smoothness(self, points: np.ndarray, min_l: float, max_l: float):
        tangents = np.diff(points, axis=0)

        scores = [
            self._angle_smoothness(tangents),
            self.normalized_length(points, min_l, max_l),
        ]

        return np.mean(scores)

    def __call__(self, segments: np.ndarray):
        if segments.shape[1] != 4:
            TardisError(
                "145",
                "tardis_em.utils.spline_metric.py",
                f"Not segmented array. Expected shape 4 got {segments.shape[1]}",
            )
        unique_ids = np.unique(segments[:, 0])
        min_l = total_length(segments[np.where(segments[:, 0] == 0)[0], 1:])
        max_l = total_length(
            segments[np.where(segments[:, 0] == segments[-1, 0])[0], 1:]
        )

        scores = []
        for i in unique_ids:
            points = segments[np.where(segments[:, 0] == i)[0], 1:]
            scores.append(self.combined_smoothness(points, min_l, max_l))

        return scores



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from os import getcwd, mkdir
from os.path import isdir, join
from shutil import rmtree
from typing import Union

import numpy as np
import torch
from torch import optim

from tardis_em.utils.logo import print_progress_bar, TardisLogo
from tardis_em.utils.utils import EarlyStopping


class ISR_LR:
    """
    Costume Inverse Square Root Learning Rate Schedular
    """

    def __init__(
        self, optimizer: optim.Adam, lr_mul: float, warmup_steps=1000, scale=100
    ):
        self._optimizer = optimizer
        self.lr_mul = lr_mul
        self.warmup_steps = warmup_steps
        self.steps = 0
        self.scale = scale
        self.param_groups = self._optimizer.param_groups

    def load_state_dict(self, checkpoint: dict):
        """
        Wrapper for loading Optimizer state dictionary

        Args:
            checkpoint (dict): Dictionary with optimizer state.
        """
        self._optimizer.load_state_dict(checkpoint)

    def state_dict(self):
        """Wrapper for retrieving Optimizer state dictionary"""
        return self._optimizer.state_dict()

    def step(self):
        """Step with the inner optimize"""
        self._update_learning_rate()
        self._optimizer.step()

    def zero_grad(self):
        """Zero out the gradients with the inner optimizer"""
        self._optimizer.zero_grad(set_to_none=True)

    def get_lr_scale(self):
        """Compute scaler for LR"""
        n_steps, n_warmup_steps = self.steps, self.warmup_steps
        return (self.scale**-0.5) * min(
            n_steps ** (-0.5), n_steps * n_warmup_steps ** (-1.5)
        )

    def _update_learning_rate(self):
        """Learning rate scheduling per step"""

        self.steps += 1
        lr = self.lr_mul * self.get_lr_scale()

        for g in self._optimizer.param_groups:
            g["lr"] = lr


class BasicTrainer:
    """
    BASIC MODEL TRAINER FOR DIST AND CNN

    Args:
        model (nn.Module): ML model build with nn.Module or nn.sequential.
        structure (dict): Model structure as dictionary.
        device (torch.device): Device for training.
        criterion (nn.loss): Loss function type.
        optimizer (optim.Adam, ISR_LR): Optimizer type.
        training_DataLoader (torch.DataLoader): DataLoader with training dataset.
        validation_DataLoader (torch.DataLoader, optional): DataLoader with test dataset.
        print_setting (tuple): Model property to display in TARDIS progress bar.
        lr_scheduler (bool): Optional Learning rate schedular.
        epochs (int): Max number of epoch's.
        early_stop_rate (int): Number of epoch's without improvement after which
            Trainer stop training.
        checkpoint_name (str): Name of the checkpoint.
    """

    def __init__(
        self,
        model,
        structure: dict,
        device: torch.device,
        criterion,
        optimizer: Union[ISR_LR, optim.Adam],
        print_setting: tuple,
        training_DataLoader,
        validation_DataLoader=None,
        lr_scheduler=False,
        epochs=100,
        early_stop_rate=10,
        instance_cov=2,
        checkpoint_name="DIST",
        classification=False,
    ):
        super(BasicTrainer, self).__init__()

        self.early_stopping = None
        self.model = model.to(device)
        self.device = device
        self.criterion = criterion
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        if lr_scheduler:
            self.lr = 1.0
        else:
            self.lr = self.optimizer.param_groups[0]["lr"]
        self.epochs = epochs
        self.early_stop_rate = early_stop_rate
        self.checkpoint_name = checkpoint_name
        self.structure = structure
        self.instance_cov = instance_cov

        if "cnn_type" in self.structure:
            self.classification = classification
            self.nn_name = self.structure["cnn_type"]
        elif "dist_type" in self.structure:
            self.nn_name = self.structure["dist_type"]

            if "node_input" in structure:
                self.node_input = structure["node_input"]

        self.training_DataLoader = training_DataLoader
        self.validation_DataLoader = validation_DataLoader

        # Set-up progress bar
        self.progress_epoch = TardisLogo()
        self.progress_train = TardisLogo()
        self.print_setting = print_setting

        self.id = 0
        self.epoch_desc = ""
        self.gpu_info = ""

        # Storage for all training metrics
        self.training_loss = []
        self.validation_loss = []
        self.learning_rate = []

        # Storage for all evaluation metrics
        self.accuracy = []
        self.precision = []
        self.recall = []
        self.f1 = []
        self.threshold = []

    @staticmethod
    def _update_desc(stop_count: int, metric: list) -> str:
        """
        Utility function to update progress bar description.

        Args:
            stop_count (int): Early stop count.
            metric (list): Best f1 score.

        Returns:
            str: Updated progress bar status.
        """
        desc = (
            f"Epochs: early_stop: {stop_count}; "
            f"F1: [{metric[0]:.2f}; {metric[1]:.2f}]"
        )
        return desc

    def _update_epoch_desc(self):
        # For each Epoch load be t model from previous run
        if self.id == 0:
            self.epoch_desc = "Epochs: early_stop: 0; best F1: NaN"
        else:
            self.epoch_desc = self._update_desc(
                self.early_stopping.counter,
                [
                    np.max(self.f1) if len(self.f1) > 0 else 0.0,
                    self.f1[-1:][0] if len(self.f1) > 0 else 0.0,
                ],
            )

    def _update_progress_bar(self, loss_desc: str, idx: int, train=True, task=""):
        """
        Update entire Tardis progress bar.

        Args:
            loss_desc (str): Description for loss function current state.
            idx (int): Number of the current epoch step.
            train (bool): If true, count progressbar for training dataset, else
                progressbar for validation
        """
        if train:
            data_set_len = len(self.training_DataLoader)
        else:
            data_set_len = len(self.validation_DataLoader)

        self.progress_train(
            title=f"{self.checkpoint_name} training module",
            text_1=self.print_setting[0],
            text_2=self.print_setting[1],
            text_3=self.print_setting[2],
            text_4=self.print_setting[3],
            text_6=task,
            text_7=self.epoch_desc,
            text_8=print_progress_bar(self.id, self.epochs),
            text_9=loss_desc,
            text_10=print_progress_bar(idx, data_set_len),
        )

    def _save_metric(self) -> bool:
        """Save training metrics"""
        if len(self.training_loss) > 0:
            np.savetxt(
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    "training_losses.csv",
                ),
                self.training_loss,
                delimiter=",",
            )

        if len(self.validation_loss) > 0:
            np.savetxt(
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    "validation_losses.csv",
                ),
                self.validation_loss,
                delimiter=",",
            )

        if len(self.f1) > 0:
            np.savetxt(
                join(getcwd(), f"{self.checkpoint_name}_checkpoint", "eval_metric.csv"),
                np.column_stack(
                    [
                        self.accuracy,
                        self.precision,
                        self.recall,
                        self.threshold,
                        self.f1,
                    ]
                ),
                delimiter=",",
            )

        if len(self.learning_rate) > 0:
            np.savetxt(
                join(
                    getcwd(), f"{self.checkpoint_name}_checkpoint", "learning_rate.csv"
                ),
                self.learning_rate,
                delimiter=",",
            )

        """ Save current model weights"""
        # If mean evaluation loss is higher than save checkpoint
        if all(self.f1[-1:][0] >= i for i in self.f1[:-1]):
            torch.save(
                {
                    "model_struct_dict": self.structure,
                    "model_state_dict": self.model.state_dict(),
                    "optimizer_state_dict": self.optimizer.state_dict(),
                },
                join(
                    getcwd(),
                    f"{self.checkpoint_name}_checkpoint",
                    f"{self.checkpoint_name}_checkpoint.pth",
                ),
            )

        torch.save(
            {
                "model_struct_dict": self.structure,
                "model_state_dict": self.model.state_dict(),
                "optimizer_state_dict": self.optimizer.state_dict(),
            },
            join(getcwd(), f"{self.checkpoint_name}_checkpoint", "model_weights.pth"),
        )

        if self.early_stopping.early_stop:
            return True
        return False

    def _mid_training_eval(self, idx):
        if idx % (len(self.training_DataLoader) // 4) == 0:
            if idx != 0 or self.id != 0:  # do not compute at trainer initialization
                # Do not validate at first idx and last 10%
                if idx != 0 and idx <= int(len(self.training_DataLoader) * 0.75):
                    self.model.eval()  # Enter Validation
                    self._validate()

                    self._update_epoch_desc()
                    self._save_metric()

                    self.model.train()  # Move back to training

    def run_trainer(self):
        """
        Main training loop.
        """
        # Initialize progress bar.
        self.progress_epoch(
            title=f"{self.checkpoint_name} training module.",
            text_2="Epoch: 0",
            text_3=print_progress_bar(0, self.epochs),
        )

        # Initialize early stop check.
        self.early_stopping = EarlyStopping(patience=self.early_stop_rate)

        # Build training directory.
        if isdir(f"{self.checkpoint_name}_checkpoint"):
            rmtree(f"{self.checkpoint_name}_checkpoint")
            mkdir(f"{self.checkpoint_name}_checkpoint")
        else:
            mkdir(f"{self.checkpoint_name}_checkpoint")

        for id_ in range(self.epochs):
            """Initialized training"""
            self.id = id_

            self._update_epoch_desc()
            self.progress_epoch(
                title=f"{self.checkpoint_name} training module",
                text_1=self.print_setting[0],
                text_2=self.print_setting[1],
                text_3=self.print_setting[2],
                text_4=self.print_setting[3],
                text_7=self.epoch_desc,
                text_8=print_progress_bar(self.id, self.epochs),
            )

            """Validation block"""
            if self.validation_DataLoader is not None and self.id != 0:
                self.model.eval()
                self._validate()

            """Training block"""
            self.model.train()
            self._train()

            stop = self._save_metric()
            if stop:
                break

    def _train(self):
        pass

    def _validate(self):
        pass



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Union

import numpy as np
import torch
from sklearn.metrics import auc, average_precision_score, roc_curve


def compare_dict_metrics(last_best_dict: dict, new_dict: dict) -> bool:
    """
    Compares two metric dictionaries and returns the one with the highest
    average metric values.

    Args:
        last_best_dict (dict): The previous best metric dictionary.
        new_dict (dict): The new metric dictionary to compare.

    Returns:
        bool: True if the new dictionary has a higher average metric value.
    """
    compare_dict_metric = lambda metric_dict: sum(metric_dict.values()) / len(
        metric_dict
    )

    last_best_dict = compare_dict_metric(last_best_dict)
    new_dict = compare_dict_metric(new_dict)

    # Compare the average metric values and return the result
    return new_dict > last_best_dict


def eval_graph_f1(
    logits: torch.Tensor, targets: torch.Tensor, threshold: float, soft=False
):
    """
    Module used for calculating training metrics

    Works with torch a numpy dataset.

    Args:
        logits (np.ndarray, torch.Tensor): Prediction output from the model.
        targets (np.ndarray, torch.Tensor): Ground truth mask.
        threshold (float):
        soft:
    """
    """Mask Diagonal as TP"""
    g_len = logits.shape[1]

    if soft:
        logits = torch.flatten(logits)
        targets = torch.flatten(targets)

        tp = torch.sum(logits * targets)
        fp = torch.sum(logits * (1 - targets))
        fn = torch.sum((1 - logits) * targets)
        tn = torch.sum((1 - logits) * (1 - targets))

        soft_prec1 = 1 - (tp / (tp + fp + 1e-16))
        soft_prec0 = 1 - (tn / (tn + fp + 1e-16))

        soft_rec1 = 1 - (tp / (tp + fn + 1e-16))
        soft_rec0 = 1 - (tn / (tn + fn + 1e-16))

        soft_f1_class1 = 1 - (2 * tp / (2 * tp + fn + fp + 1e-16))
        soft_f1_class0 = 1 - (2 * tn / (2 * tn + fn + fp + 1e-16))

        prec_cost = torch.mean(0.5 * (soft_prec1 + soft_prec0))
        rec_cost = torch.mean(0.5 * (soft_rec1 + soft_rec0))
        f1_cost = torch.mean(0.5 * (soft_f1_class1 + soft_f1_class0))
        return prec_cost, rec_cost, f1_cost
    else:
        """Find best f1 based on variation threshold"""
        f1 = []
        acc = []
        precision = []
        recall = []

        if threshold == 0.0:
            while threshold != 1:
                input_df = torch.where(logits > threshold, 1, 0)

                tp, fp, tn, fn = confusion_matrix(input_df, targets)
                tp = tp - len(input_df)  # remove diagonal from F1

                accuracy_score = (tp + tn) / (tp + tn + fp + fn + 1e-16)
                prec = tp / (tp + fp + 1e-16)
                rec = tp / (tp + fn + 1e-16)

                acc.append(accuracy_score)
                precision.append(prec)
                recall.append(rec)
                f1.append(2 * (prec * rec) / (prec + rec + 1e-16))

                threshold = round(threshold + 0.01, 2)

            threshold = np.where(f1 == np.max(f1))[0]

            if threshold.shape[0] > 1:
                th = 0

                for i in threshold:
                    th = th + round(i * 0.01, 2)

                threshold = round(th / len(threshold), 2)
            else:
                threshold = round(float(threshold) * 0.01, 2)

            id_ = int(threshold * 100)
            accuracy_score = round(acc[id_], 2)
            precision_score = round(precision[id_], 2)
            recall_score = round(recall[id_], 2)
            F1_score = round(f1[id_], 2)
        else:
            accuracy_score, precision_score, recall_score, F1_score = 0, 0, 0, 0
            input_df = torch.where(logits >= threshold, 1, 0)
            idx_1 = torch.where(targets == 1)
            idx_0 = torch.where(targets == 0)

            # IDX_1
            input_ = input_df[idx_1]
            target_ = targets[idx_1]

            tp, fp, tn, fn = confusion_matrix(input_, target_)
            tp = tp - g_len

            accuracy_score += (tp + tn) / (tp + tn + fp + fn + 1e-16)
            precision_score += tp / (tp + fp + 1e-16)
            recall_score += tp / (tp + fn + 1e-16)
            F1_score += (
                2
                * (precision_score * recall_score)
                / (precision_score + recall_score + 1e-16)
            )

            # IDX_0
            input_ = input_df[idx_0]
            target_ = targets[idx_0]

            tn, fn, tp, fp = confusion_matrix(input_, target_)

            acc = (tp + tn) / (tp + tn + fp + fn + 1e-16)
            accuracy_score += acc
            prec = tp / (tp + fp + 1e-16)
            precision_score += prec
            rec = tp / (tp + fn + 1e-16)
            recall_score += rec
            F1_score += 2 * (prec * rec) / (prec + rec + 1e-16)

        return (
            accuracy_score / 2,
            precision_score / 2,
            recall_score / 2,
            F1_score / 2,
            threshold,
        )


def calculate_f1(
    logits: Union[np.ndarray, torch.Tensor],
    targets: Union[np.ndarray, torch.Tensor],
    best_f1=True,
):
    """
    Module used for calculating training metrics

    Works with torch a numpy dataset.

    Args:
        logits (np.ndarray, torch.Tensor): Prediction output from the model.
        targets (np.ndarray, torch.Tensor): Ground truth mask.
        best_f1 (bool): If True an expected inputs is probability of classes and
            measured metrics is soft-f1.
    """
    """Find best f1 based on variation threshold"""
    if best_f1:
        threshold = 0
        f1 = []
        acc = []
        precision = []
        recall = []

        while threshold != 1:
            input_df = torch.where(logits > threshold, 1, 0)

            tp, fp, tn, fn = confusion_matrix(input_df, targets)

            accuracy_score = (tp + tn) / (tp + tn + fp + fn + 1e-16)
            prec = tp / (tp + fp + 1e-16)
            rec = tp / (tp + fn + 1e-16)

            acc.append(accuracy_score)
            precision.append(prec)
            recall.append(rec)
            f1.append(2 * (prec * rec) / (prec + rec + 1e-16))

            threshold = round(threshold + 0.01, 2)

        threshold = np.where(f1 == np.max(f1))[0]
        if threshold.shape[0] > 1:
            th = 0

            for i in threshold:
                th = th + round(i * 0.01, 2)

            threshold = round(th / len(threshold), 2)
        else:
            threshold = round(float(threshold) * 0.01, 2)

        id_ = int(threshold * 100)
        accuracy_score = round(acc[id_], 2)
        precision_score = round(precision[id_], 2)
        recall_score = round(recall[id_], 2)
        F1_score = round(f1[id_], 2)

        return accuracy_score, precision_score, recall_score, F1_score, threshold
    else:
        tp, fp, tn, fn = confusion_matrix(logits, targets)

        """Accuracy Score - (tp + tn) / (tp + tn + fp + fn)"""
        accuracy_score = (tp + tn) / (tp + tn + fp + fn + 1e-16)

        """Precision Score - tp / (tp + fp)"""
        precision_score = tp / (tp + fp + 1e-16)

        """Recall Score - tp / (tp + tn)"""
        recall_score = tp / (tp + fn + 1e-16)

        """F1 Score - 2 * [(Prec * Rec) / (Prec + Rec)]"""
        F1_score = 2 * (
            (precision_score * recall_score) / (precision_score + recall_score + 1e-16)
        )

        return accuracy_score, precision_score, recall_score, F1_score


def AP(logits: np.ndarray, targets: np.ndarray) -> float:
    logits = logits.flatten()
    targets = targets.flatten()

    return average_precision_score(targets, logits)


def AP_instance(input_: np.ndarray, targets: np.ndarray) -> float:
    prec = 0

    # Get GT instances, compute IoU for best mache between GT and input
    for j in np.unique(targets[:, 0]):
        true_c = targets[np.where(targets[:, 0] == j)[0], 1:]  # Pick GT instance
        prec_df = []

        # Select max Prec (best mach)
        for i in np.unique(input_[:, 0]):
            pred = input_[np.where(input_[:, 0] == i)[0]]  # Pick input instance

            # Prec is ratio of true positives to the total number of positive detections
            prec_df.append(
                sum([True for i in true_c if i in pred[:, 1:]]) / len(true_c)
            )
        prec += np.max(prec_df)

    return 1 / (len(np.unique(targets[:, 0]) + 1e-16)) * prec


def AUC(logits: np.ndarray, targets: np.ndarray, diagonal=False) -> float:
    if diagonal:
        g_len = logits.shape[1]
        g_range = range(g_len)

        if logits.ndim == 3:
            logits[:, g_range, g_range] = 1.0
            targets[:, g_range, g_range] = 1.0
        else:
            logits[g_range, g_range] = 1.0
            targets[g_range, g_range] = 1.0

    logits = logits.flatten()
    targets = targets.flatten()

    fpr, tpr, _ = roc_curve(targets, logits)
    return auc(fpr, tpr)


def IoU(input_: np.ndarray, targets: np.ndarray, diagonal=False):
    if diagonal:
        g_len = input_.shape[1]
        g_range = range(g_len)

        if input_.ndim == 3:
            input_[:, g_range, g_range] = 1.0
            targets[:, g_range, g_range] = 1.0
        else:
            input_[g_range, g_range] = 1.0
            targets[g_range, g_range] = 1.0

    input_ = input_.flatten()
    targets = targets.flatten()

    tp, fp, tn, fn = confusion_matrix(input_, targets)

    return tp / (tp + fp + fn + 1e-16)


def mcov(
    input_,
    targets,
):
    """
    Mean Coverage metric

    Args:
        input_ (np.ndarray, torch.Tensor): _description_
        targets (np.ndarray, torch.Tensor): _description_

    Returns:
        float: _description_
    """
    mCov = 0
    mwCov = 0

    unique_target = np.unique(targets[:, 0])
    G = len(unique_target)
    unique_input = np.unique(input_[:, 0])

    # Get GT instances, compute IoU for best mach between GT and input
    for j in unique_target:
        g = targets[targets[:, 0] == j, 1:]  # Pick GT instance
        w_g = g.shape[0] / targets.shape[0]  # ratio of instance to whole PC
        iou = []

        # Select max IoU (the best mach)
        for i in unique_input:
            p = input_[input_[:, 0] == i, 1:]  # Pick input instance

            # Intersection of coordinates between GT and input instances
            # Union of coordinates between GT and input instances

            intersection = np.sum(np.any(np.isin(p, g), axis=1))
            if intersection > 0:
                union = np.unique(np.concatenate((p, g), axis=0), axis=0).shape[0]
                iou.append(intersection / union)

        if len(iou) > 0:
            max_iou = np.max(iou)
        else:
            max_iou = 0.0

        mwCov += w_g * max_iou
        mCov += max_iou

    return mCov / G, mwCov


def confusion_matrix(
    logits: Union[np.ndarray, torch.Tensor],
    targets: Union[np.ndarray, torch.Tensor],
):
    if torch.is_tensor(logits):
        confusion_vector = logits / targets

        tp = torch.sum(confusion_vector == 1).item()
        fp = torch.sum(confusion_vector == float("inf")).item()
        tn = torch.sum(torch.isnan(confusion_vector)).item()
        fn = torch.sum(confusion_vector == 0).item()
    else:
        logits = normalize_image(logits)
        targets = normalize_image(targets)

        with np.errstate(divide="ignore", invalid="ignore"):
            confusion_vector = logits / targets

        tp = np.sum(confusion_vector == 1)
        fp = np.sum(confusion_vector == float("inf"))
        tn = np.sum(np.isnan(confusion_vector) is True)
        fn = np.sum(confusion_vector == 0)

    return tp, fp, tn, fn


def normalize_image(image: np.ndarray):
    """
    Simple image data normalizer between 0,1

    Args:
        image (np.ndarray): Image data set.
    """
    image_min = np.min(image)
    image_max = np.max(image)

    if image_min == 0 and image_max == 1:
        return image

    if image_min == 0:
        image = np.where(image > image_min, 1, 0)
    elif image_max == 0:
        image = np.where(image < image_max, 1, 0)

    return image
